@@ -1,6 +1,19 @@
+from collections import defaultdict
 from dataclasses import dataclass
 from itertools import repeat
-from typing import Any, Hashable, Iterable, Literal, Mapping, Sequence, Union, overload
+from typing import (
+    Any,
+    Collection,
+    Hashable,
+    Iterable,
+    KeysView,
+    Literal,
+    Mapping,
+    Sequence,
+    Sized,
+    Union,
+    overload,
+)
 
 import numpy as np
 from numpy import typing as npt
@@ -296,6 +309,7 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
         columns: Iterable[Hashable] | None = None,
         index: Any = None,
         copy: bool = False,
+        dtype: Mapping[Hashable, npt.DTypeLike] | None = None,
         check: bool = True,
     ):
         """
@@ -308,17 +322,20 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
         :param columns: Override column names. May not be specified if `data` is a mapping.
         :param index: Index column keys or values.
         :param copy: Do not copy any arrays unless have to.
+        :param dtype: Mapping from column names to the desired dtypes.
         :param check: Validate the integrity after constructing the instance.
         """
+        if dtype is None:
+            dtype = {}
         if data is None:
             if columns is None:
                 built_columns = {}
             else:
-                built_columns = {c: np.array([], dtype=object) for c in columns}
+                built_columns = {c: np.array([], dtype=dtype.get(c, object)) for c in columns}
         elif isinstance(data, Mapping):
             if columns is not None:
                 raise ValueError("Either `data` or `columns` must be specified")
-            built_columns = {k: np.array(v, copy=copy) for k, v in data.items()}
+            built_columns = {k: self._asarray(v, dtype.get(k), copy) for k, v in data.items()}
         else:
             if columns is None:
                 columns = repeat(None)
@@ -326,7 +343,7 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
             for i, (arr, name) in enumerate(zip(data, columns)):
                 if name is None:
                     name = str(i)
-                built_columns[name] = np.array(arr, copy=copy)
+                built_columns[name] = self._asarray(arr, dtype.get(name), copy)
         self._columns: dict[Hashable, np.ndarray] = built_columns
         if check:
             self._check_columns(built_columns)
@@ -348,26 +365,30 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
         ...  # pragma: no cover
 
     @overload
-    def __getitem__(self, item: list[Hashable]) -> "DataFrame":  # noqa: D105
+    def __getitem__(self, item: Collection[Hashable]) -> "DataFrame":  # noqa: D105
         ...  # pragma: no cover
 
-    def __getitem__(self, item: Hashable | list[Hashable]) -> Union[np.ndarray, "DataFrame"]:
+    def __getitem__(self, item: Hashable | Collection[Hashable]) -> Union[np.ndarray, "DataFrame"]:
         """
         Extract the column values by key. Raise `KeyError` if the column doesn't exist.
 
-        If the item is an array, return the DataFrame with the corresponding columns.
+        If the item is a collection of keys, return the DataFrame with the corresponding columns.
 
         :param item: Column key.
         :return: Column values, zero-copy. The user is welcome to mutate the internals.
         """
-        if isinstance(item, list):
-            item = item.copy()
+        scalar = isinstance(item, Hashable)
+        if scalar and isinstance(item, (tuple, frozenset)) and item not in self._columns:
+            scalar = False
+        if not scalar:
+            assert isinstance(item, Iterable)
+            keys = list(item)
             columns = self._columns
             for i in self._index:
-                if i not in item:
-                    item.append(i)
-            return DataFrame({k: columns[k] for k in item}, index=self._index)
-        return self._columns[item]
+                if i not in keys:
+                    keys.append(i)
+            return DataFrame({k: columns[k] for k in keys}, index=self._index)
+        return self._columns[item]  # type: ignore
 
     def __setitem__(self, key: Hashable, value: Any) -> None:
         """
@@ -396,12 +417,18 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
 
         :param key: Column key to delete.
         """
+        if key in self._index:
+            raise ValueError("cannot drop a part of the index, please use set_index()")
         del self._columns[key]
 
     def __contains__(self, item: Hashable) -> bool:
         """Check whether the column key is present."""
         return item in self._columns
 
+    def __iter__(self) -> Iterable[Hashable]:
+        """Iterate over columns."""
+        return iter(self._columns.keys())
+
     def __str__(self) -> str:
         """Support str()."""
         # fmt: off
@@ -851,7 +878,7 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
     def isin(
         self,
         column: Hashable,
-        haystack: npt.ArrayLike,
+        haystack: npt.ArrayLike | set | frozenset | dict | KeysView,
         assume_unique: bool = False,
         invert: bool = False,
     ) -> npt.NDArray[np.bool_]:
@@ -866,6 +893,9 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
         :return: Boolean mask, same shape as `column`.
         """
         values = self[column]
+        if isinstance(haystack, (set, frozenset, dict, KeysView)):
+            # do not shoot in the foot
+            haystack = _NumpyIterableProxy(haystack)
         haystack = np.asarray(haystack, dtype=values.dtype)
         kind = values.dtype.kind
         if kind == "S" or kind == "U":
@@ -1018,6 +1048,7 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
         *dfs: "DataFrame",
         ignore_index: bool = False,
         copy: bool = False,
+        strict: bool = True,
     ) -> "DataFrame":
         if len(dfs) == 0:
             return cls()
@@ -1029,24 +1060,44 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
             return dfs[0].copy()
         index = dfs[0]._index
         columns = dfs[0]._columns.keys()
-        empty = dfs[0].empty
-        concat_columns = {k: [v] if not empty else [] for k, v in dfs[0]._columns.items()}
-        for df in dfs[1:]:
+        column_chunks = defaultdict(list)
+        empty_column_dtypes: dict[Hashable, np.dtype] = {}
+        pos = 0
+        for df in dfs:
             if not isinstance(df, cls):
                 raise TypeError(f"Can only concatenate medvedi.DataFrame-s, got {type(df)}")
-            if df._columns.keys() != columns:
+            if strict and df._columns.keys() != columns:
                 raise ValueError(f"Columns must match: {columns} vs. {df._columns.keys()}")
             if df._index != index and not ignore_index:
                 raise ValueError(f"Indexes must match: {index} vs. {df._index}")
-            if len(df):
+            if df_len := len(df):
+                arange = None if strict else np.arange(pos, pos + df_len)
+                pos += df_len
                 for k, v in df._columns.items():
-                    concat_columns[k].append(v)
+                    column_chunks[k].append((v, arange))
+            else:
+                for k, v in df._columns.items():
+                    empty_column_dtypes.setdefault(k, v.dtype)
+                    column_chunks.setdefault(k, [])
+        concat_columns = {}
+        for key, chunks in column_chunks.items():
+            if not chunks:
+                column = cls._empty_array(pos, empty_column_dtypes[key])
+            else:
+                values = np.concatenate([arr for arr, _ in chunks], casting="unsafe")
+                if strict:
+                    column = values
+                else:
+                    indexes = np.concatenate([i for _, i in chunks])
+                    if len(indexes) == pos:
+                        column = values
+                    else:
+                        column = cls._empty_array(pos, values.dtype)
+                        column[indexes] = values
+            concat_columns[key] = column
         concat_df = cls()
         concat_df._index = index if not ignore_index else ()
-        concat_df._columns = {
-            k: np.concatenate(v, casting="unsafe") if len(v) else dfs[0]._columns[k]
-            for k, v in concat_columns.items()
-        }
+        concat_df._columns = concat_columns
         return concat_df
 
     @classmethod
@@ -1240,6 +1291,15 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
             return np.full(length, None, dtype)
         return np.zeros(length, dtype=dtype)
 
+    @staticmethod
+    def _asarray(value: Sequence | np.ndarray, dtype: npt.DTypeLike, copy: bool) -> np.ndarray:
+        if dtype == object and not isinstance(value, np.ndarray):
+            r = np.empty(len(value), dtype)
+            r[:] = value
+        else:
+            r = np.array(value, dtype, copy=copy)
+        return r
+
     @staticmethod
     def _check_columns(columns: dict[Hashable, np.ndarray]) -> None:
         length = None
@@ -1250,3 +1310,20 @@ class DataFrame(metaclass=PureStaticDataFrameMethods):
                 raise ValueError(f"all columns must have equal length: {k}")
             if v.ndim != 1:
                 raise ValueError(f"column {k} must be 1-dimensional, got {v.ndim}")
+
+
+class _NumpyIterableProxy(Sequence[Any]):
+    __slots__ = ("obj",)
+
+    def __init__(self, obj):
+        assert isinstance(obj, (Iterable, Sized))
+        self.obj = obj
+
+    def __len__(self):
+        return len(self.obj)
+
+    def __getitem__(self, item):
+        raise AssertionError("numpy shouldn't call this")
+
+    def __iter__(self):
+        return iter(self.obj)
